from __future__ import annotations

import logging
import typing
from inspect import signature, Signature, getdoc, cleandoc, Parameter
from types import MappingProxyType

logger = logging.getLogger(__name__)


class EndpointNotFound(Exception):
    """Raised if endpoint is not found by name."""


class Endpoint:
    def __init__(self, name: str, handler: typing.Callable):
        self.name = name
        self.handler = handler
        self.signature = signature(self.handler)
        self.enabled = True
        self.check_annotations()

    def __str__(self) -> str:
        return f"Endpoint('{self.name}')"

    def __call__(self, *args, **kwargs):
        return self.handler(*args, **kwargs)

    @property
    def doc(self) -> str:
        doc = getdoc(self.handler)
        return cleandoc(doc) if doc else ""

    @property
    def parameters(self) -> MappingProxyType[str, Parameter]:
        return self.signature.parameters

    def get_parameter(self, name: str) -> Parameter:
        return self.parameters[name]

    def check_annotations(self):
        for parameter in self.parameters.values():
            if parameter.annotation is Signature.empty:
                logger.warning(
                    "%s() takes '%s' as an argument but is missing type annotation. "
                    "Server considers the argument as 'str' unless client call provides "
                    "the missing type annotation.",
                    self.name,
                    parameter.name,
                )

    def to_dict(self) -> dict:
        logger.debug("Converting '%s' parameters to dict.", self)

        parameters = [self._parameter_to_dict(p) for p in self.parameters.values()]
        definition = {"name": self.name, "documentation": self.doc, "parameters": parameters}

        if self.signature.return_annotation is not Signature.empty:
            return_annotation = annotation_to_str(self.signature.return_annotation)
        else:
            return_annotation = ""

        definition["return_annotation"] = return_annotation

        return definition

    @classmethod
    def _parameter_to_dict(cls, parameter):
        logger.debug("Converting '%s' parameter to dict.", parameter)

        definition = {
            "name": get_name(parameter),
            "annotation": get_annotation(parameter),
        }

        if parameter.default is not Parameter.empty:
            definition["default"] = get_default(parameter)

        return definition


def is_optional(parameter: Parameter) -> bool:
    """Is parameter optional.

    Returns True if a parameter is optional.

    For example,

        def sample(a: Optional[int] = 1):
            pass
    """
    origin = typing.get_origin(parameter.annotation)
    return origin is typing.Union and type(None) in typing.get_args(parameter.annotation)


def get_name(parameter: Parameter) -> str:
    return parameter.name


def get_annotation(parameter: Parameter) -> str:
    annotation = parameter.annotation

    if annotation is Parameter.empty:
        return type(parameter.default).__name__ if parameter.default else ""

    if is_optional(parameter):
        return str(annotation)

    return annotation_to_str(annotation)


def annotation_to_str(annotation) -> str:
    if isinstance(annotation, str):
        return annotation
    return annotation.__name__


def get_default(parameter: Parameter) -> str:
    return repr(parameter.default)


"""NumPy codec to be used with remotecall library.

Usage:
    from remotecall.extracodecs.imagecodec import NumPyCodec
"""
from __future__ import annotations

import typing

import numpy as np
from numpy.typing import NDArray, DTypeLike

from remotecall import Codec


class NumPyCodec(Codec):
    """NumPy codec.

    Example NumPy codec.
    """

    @classmethod
    def is_matching_content_type(cls, content_type: str) -> bool:
        return content_type.startswith("application/numpy")

    @classmethod
    def is_matching_type(cls, type_: typing.Type) -> bool:
        return type_ == np.ndarray

    @classmethod
    def create_by_value(cls, array: NDArray) -> NumPyCodec:
        return cls(shape=array.shape, dtype=array.dtype)

    @classmethod
    def create_by_content_type(cls, content_type: str) -> NumPyCodec:
        shape, data_type = cls._get_shape_and_data_type(content_type)
        return cls(shape=shape, dtype=data_type)

    @classmethod
    def _get_shape_and_data_type(cls, content_type: str) -> tuple[list, DTypeLike]:
        """Extract shape and dtype from content-type.

        Content-type string is expected to be like "numpy-uint8-1080x1920x3".
        """
        fields = content_type.split("-")
        return (
            [int(dimension) for dimension in fields[2].split("x")],
            np.dtype(fields[1])
        )

    def __init__(self, shape: list, dtype: DTypeLike):
        self.shape = shape
        self.dtype = dtype

    @classmethod
    def get_type(cls) -> typing.Type:
        return np.ndarray

    def encode(self, array: NDArray) -> bytes:
        return array.tobytes()

    def decode(self, data: bytes) -> NDArray:
        return np.ndarray(shape=self.shape, dtype=self.dtype, buffer=data)

    def get_content_type(self) -> str:
        # Example output: "numpy-uint8-1080x1920x3"
        dtype_name = str(self.dtype)
        shape_name = "x".join(map(str, self.shape))
        return f"application/numpy-{dtype_name}-{shape_name}"



"""
Provides functionality to fetch API definition from a server and generate client code.

Usage:

    Fetch API from server:

        $ python -m remotecall fetch_api http://localhost:8000 --output api.json

    Generate client from JSON:

        $ python -m remotecall generate_client api.json --output client.py

    Generate client from URL:

        $ python -m remotecall generate_client http://localhost:8000 --output client.py
"""
from __future__ import annotations

import argparse
import json
import logging
import os
import typing

from . import fetch_api
from . import generate_client


logger = logging.getLogger(__name__)


DEFAULT_INDENT = 4
DEFAULT_TIMEOUT = 5


def _handle_fetch_api(
        server_url: str,
        cert_file: typing.Optional[str] = None,
        output_file: typing.Optional[str] = None,
        indent: int = DEFAULT_INDENT,
        timeout: int = DEFAULT_TIMEOUT):
    """Handler for fetching API definition."""
    definition = fetch_api(url=server_url, cert_file=cert_file, timeout=timeout)

    # JSON library transforms data to valid ascii objects by default (ensure_ascii=True). Thus,
    # non-ASCII characters are ASCII encoded (e.g., 'Ä' is \u00c4).

    if not output_file:
        print(json.dumps(definition, indent=indent, ensure_ascii=False))
        return

    with open(output_file, "w", encoding="utf-8") as fh:
        json.dump(definition, fh, indent=indent, ensure_ascii=True)


def _handle_generate_client(source: str,
                            cert_file: typing.Optional[str] = None,
                            output_file: typing.Optional[str] = None,
                            class_name: typing.Optional[str] = None):
    """Handler for client generation."""
    if source.startswith("http"):
        definition = fetch_api(url=source, cert_file=cert_file)
    else:
        definition = _load_from_json(filename=source)

    source_code = generate_client(definition=definition, class_name=class_name)

    if not output_file:
        print(source_code)
        return

    with open(output_file, "w", encoding="utf-8") as fh:
        fh.write(source_code)


def _load_from_json(filename: str) -> dict:
    """Load API definition from JSON file."""
    with open(filename, "rb", encoding="utf-8") as fh:
        return json.load(fh)


def main(cli_args):
    """Main"""
    parser = argparse.ArgumentParser(prog=__name__)
    parser.add_argument("command", choices=["fetch_api", "generate_client"])
    parser.add_argument("source")
    parser.add_argument("--output", help="output")
    parser.add_argument("--certificate", help="SLL certificate")
    parser.add_argument("--name", help="Client class name.")

    args = parser.parse_args(cli_args)

    cert_file = os.path.abspath(args.certificate) if args.certificate else None

    if args.command == "fetch_api":
        _handle_fetch_api(server_url=args.source,
                          cert_file=cert_file,
                          output_file=args.output)
    elif args.command == "generate_client":
        _handle_generate_client(source=args.source,
                                cert_file=cert_file,
                                output_file=args.output,
                                class_name=args.name)
    else:
        print(f"Unknown command: {args.command}")


if __name__ == "__main__":
    import sys

    main(sys.argv[1:])



from __future__ import annotations

from typing import Any
import logging
import requests

from .codecs import Codec
from .codecs import Codecs
from .constants import headers
from .errors import ClientError
from .errors import ServerError


logger = logging.getLogger(__name__)


class BaseClient:
    DEFAULT_SERVER_ADDRESS = ("localhost", 8000)

    def __init__(self, server_address: tuple[str, int] = DEFAULT_SERVER_ADDRESS):
        self.server_address = server_address
        self._codecs = Codecs(Codec.subclasses)
        self._ssl_enabled = False
        self._cert_file = None
        self._authentication = None

    @property
    def server_url(self) -> str:
        """Get server URL."""
        host, port = self.server_address
        scheme = "https" if self._ssl_enabled else "http"
        return f"{scheme}://{host}:{port}"

    @property
    def ssl_enabled(self) -> bool:
        """Is SSL enabled."""
        return self._ssl_enabled

    def use_ssl(self, cert_file: str):
        """Enable SSL."""
        self._cert_file = cert_file
        self._ssl_enabled = True

    def set_authentication(self, authentication):
        """Set authentication method."""
        self._authentication = authentication

    def call(self, function_name: str, timeout=None, **kwargs):
        """Call remote function with arguments."""
        logger.debug("Preparing to call %s.", function_name)
        url = self._get_url(function_name)
        multipart_form_data = self._encode_multipart_form_data(fields=kwargs)

        logger.debug("Sending POST request %s.", url)

        response = requests.post(url=url,
                                 files=multipart_form_data,
                                 verify=self._cert_file,
                                 auth=self._authentication,
                                 timeout=timeout)

        logger.debug("Received status code: %s.", response.status_code)

        if response.status_code >= 500:
            self._log_response(response)
            raise ServerError(f"Server error: {response.text}", response.status_code)

        if response.status_code >= 400:
            self._log_response(response)
            raise ClientError(f"Client error: {response.text}", response.status_code)

        return self._get_response_value(response)

    def _get_url(self, command: str) -> str:
        return f"{self.server_url}/{command}"

    def _encode_multipart_form_data(self, fields: dict[str, Any]) -> dict[str, tuple]:
        logger.debug("Encoding multipart/form-data ...")
        form_data = {}

        for name, value in fields.items():
            logger.debug("Get codec for '%s.", type(value))
            codec = self._codecs.get_codec_by_value(value)  # OK

            logger.debug("Encoding '%s' with %s.", name, codec)
            form_data[name] = (None, codec.encode(value), codec.get_content_type())

        return form_data

    @classmethod
    def _log_response(cls, response: requests.Response):
        logger.debug("Response:")
        logger.debug("  status: %s", response.status_code)
        logger.debug("  headers: %s", response.headers)
        logger.debug("  text: %s", response.text)

    def _get_response_value(self, response: requests.Response):
        content_type = response.headers.get(str(headers.CONTENT_TYPE))

        if not content_type:
            return response.content

        codec = self._codecs.get_codec_by_content_type(content_type)

        logger.debug("Using %s to decode '%s' response.", codec, content_type)

        return codec.decode(response.content)



from __future__ import annotations

import http.server
import logging
import typing
from typing import Optional
from .endpoint import Endpoint
from .requesthandler import HTTPRequestHandler
from .codecs import Codec
from .codecs import Codecs
from .authentication import Authenticator


logger = logging.getLogger(__name__)


class Server(http.server.ThreadingHTTPServer):
    """Server."""

    def __init__(self, server_address: tuple[str, int]):
        super().__init__(
            server_address=server_address, RequestHandlerClass=HTTPRequestHandler
        )
        self._ssl_enabled = False
        self._authenticator: Optional[Authenticator] = None
        self.codecs = Codecs(Codec.subclasses)
        self.endpoints: typing.Dict[str, Endpoint] = {}

    @property
    def url(self) -> str:
        """Server URL."""
        hostname, port = self.server_address
        scheme = "https" if self.ssl_enabled else "http"
        return f"{scheme}://{hostname}:{port}"

    @classmethod
    def _create_endpoint(cls, function: typing.Callable, name: str = None):
        """Create endpoint."""
        if not name:
            try:
                name = function.__name__
            except AttributeError as err:
                logger.warning("Endpoint name is optional if the value can be read from "
                               "the provided function by calling function.__name__.")
                raise ValueError("Endpoint name not provided and exposed function raised the "
                                 f"following error when calling function.__name__: {err}") from err

        return Endpoint(name, function)

    def expose(self, function: typing.Callable, name: str = None):
        """Expose a function with given name.

        Name of the callable shall be used if name is not given. The name is resolved by calling
        __name__ of the callable.
        """
        logger.debug("Exposing %s ...", function)

        endpoint = self._create_endpoint(function, name)
        self.endpoints[endpoint.name] = endpoint

        logger.debug("Exposed %s as %s", function, endpoint)

    @property
    def ssl_enabled(self) -> bool:
        """Is SLL enabled."""
        return self._ssl_enabled

    def use_ssl(self, cert_file: str, key_file: str, password=None):
        """Use SSL.

        Use SSL to secure server to client transactions.
        """
        import ssl

        logger.debug("Setting up SSL context.")
        logger.debug("  Certificate file: %s", cert_file)
        logger.debug("  Key file: %s", key_file)

        context = ssl.SSLContext(ssl.PROTOCOL_TLS_SERVER)
        context.check_hostname = False
        context.load_cert_chain(certfile=cert_file, keyfile=key_file, password=password)

        self.socket = context.wrap_socket(self.socket, server_side=True)
        self._ssl_enabled = True

        logger.debug("SSL enabled.")

    def set_authenticator(self, authenticator: Authenticator):
        """Set authenticator."""
        self._authenticator = authenticator

    def get_authenticator(self) -> Authenticator:
        """Get authenticator."""
        return self._authenticator



class ClientError(Exception):
    """4xx: Client error"""

    def __init__(self, message: str, status_code: int):
        self.message = message
        self.status_code = status_code
        super().__init__(self.message)


class ServerError(Exception):
    """5xx: Server error"""

    def __init__(self, message: str, status_code: int = 500):
        self.message = message
        self.status_code = status_code
        super().__init__(self.message)



from __future__ import annotations

import requests
from typing import Optional
from typing import Tuple
import urllib

from .errors import ClientError
from .errors import ServerError
from .server import Server
from .client import BaseClient
from .codecs import Codec
from .constants import headers
from .constants import statuscodes


def fetch_api(url: str, cert_file: Optional[str] = None, timeout: Optional[float] = None) -> dict:
    """Fetch API definition from URL."""
    hostname, port = _parse_hostname_and_port(url)
    response = requests.get(url=url, verify=cert_file, timeout=timeout)
    definition = response.json()
    definition["address"] = {"hostname": hostname, "port": port}
    return definition


def _parse_hostname_and_port(url: str) -> Tuple[str, int]:
    url_obj = urllib.parse.urlparse(url)
    return url_obj.hostname, url_obj.port


def generate_client(definition: dict, class_name: Optional[str] = None):
    from .clientfactory import ClientFactory
    factory = ClientFactory(definition, class_name=class_name)
    return factory.generate()



from __future__ import annotations

import cgi
from dataclasses import dataclass
import re
import logging
import typing
from typing import BinaryIO
from typing import Optional

from .response import Response
from .errors import ClientError
from .constants import headers
from .constants.headers import CONTENT_TYPE
from .constants.headers import CONTENT_DISPOSITION
from .constants.statuscodes import TYPE_ERROR
from .constants.statuscodes import BAD_REQUEST
from .constants.statuscodes import SERVICE_UNAVAILABLE

if typing.TYPE_CHECKING:
    from .endpoint import Endpoint
    from .requesthandler import HTTPRequestHandler


logger = logging.getLogger(__name__)


@dataclass
class MultiPartFormDataField:
    name: str
    content_type: str
    data: bytes


class PostHandler:

    def __init__(self, request: HTTPRequestHandler):
        self.request = request
        self._server = self.request.get_server()
        self._codecs = self._server.codecs

    def handle(self) -> Response:  # pylint: disable=broad-exception-caught
        endpoint_name = self.request.path_component

        try:
            endpoint = self._server.endpoints[endpoint_name]
        except KeyError as err:
            raise ClientError(f"Unknown endpoint: '{endpoint_name}'", BAD_REQUEST) from err

        if not endpoint.enabled:
            raise ClientError("Endpoint disabled.", SERVICE_UNAVAILABLE)

        try:
            arguments = self.get_arguments(endpoint)
        except ValueError as err:
            raise ClientError(
                f"Failed to parse request arguments: {err}", BAD_REQUEST
            ) from err

        try:
            return_value = endpoint(**arguments)
        except TypeError as err:
            raise ClientError(str(err), TYPE_ERROR) from err

        try:
            codec = self._codecs.get_codec_by_value(return_value)
            logger.debug("Using %s to decode return value.", codec)
            response = Response(codec.encode(return_value), codec.get_content_type())
        except Exception as err:
            raise ClientError(
                f"Failed to encode return value: {err}", TYPE_ERROR
            ) from err

        return response

    def get_arguments(self, endpoint: Endpoint) -> dict:
        logger.debug("Get arguments.")

        arguments = {}

        request = self.request
        content_length = request.content_length
        content_type_header = request.headers.get(CONTENT_TYPE)

        if not content_type_header:
            logger.debug("No arguments ('content-type' not set).")
            return arguments

        request_content_type, options = cgi.parse_header(content_type_header)
        if request_content_type != headers.MULTIPART_FORM_DATA:
            raise ClientError(
                f"Invalid content-type: '{request_content_type}'.", BAD_REQUEST,
            )

        boundary = options.get(headers.BOUNDARY)

        if not boundary:
            raise ClientError("Boundary not defined.", BAD_REQUEST)

        fields = self.read_multipart_form_data(request.rfile, content_length, boundary)

        for field in fields:
            arguments[field.name] = self.decode_multipart_form_data(field, endpoint)

        return arguments

    def decode_multipart_form_data(self, field: MultiPartFormDataField, endpoint: Endpoint):
        try:
            parameter = endpoint.get_parameter(field.name)
        except KeyError as err:
            raise ClientError(
                f"{endpoint.name}() got unexpected parameter: {field.name}", BAD_REQUEST
            ) from err

        if field.content_type:
            codec = self._codecs.get_codec_by_content_type(field.content_type)
        else:
            codec = self._codecs.get_codec_by_type(parameter.annotation or str)

        logger.debug("Using %s to decode '%s'.", codec, field.name)

        return codec.decode(field.data)

    @classmethod
    def read_multipart_form_data(
        cls, rfile: BinaryIO, length: int, boundary: str
    ) -> list[MultiPartFormDataField]:
        """
        Example:
            b'--40c00dbb9994e1de1cf94bce01c4e734'
            b'Content-Disposition: form-data; name="text"'
            b'Content-Type: application/str'
            b''
            b'Hello world'
            b'--40c00dbb9994e1de1cf94bce01c4e734--'
            b''
        """

        def read_lines(rfile: BinaryIO, length: int) -> list[bytes]:
            bytes_count = 0

            for line in rfile:
                yield line

                bytes_count += len(line)
                if bytes_count >= length:
                    break

        def get_name(text: str) -> Optional[str]:
            # Example:
            #   Content-Disposition: form-data; name="a"; filename="a"
            try:
                return re.search('name="(.+?)"', text).group(1)
            except AttributeError:
                logger.warning("Expecting name field in '%s'.", text)
                return None

        def get_content_type(text: str) -> str:
            return text.split(":")[1].strip()

        fields = []
        name = None
        content_type = None
        delimiter = b"--" + boundary.encode()
        expect_data = False
        data = []

        for line in read_lines(rfile, length):
            # logger.debug(line)

            # Note: this will also catch the end delimiter with trailing dashes ('--').
            if line.startswith(delimiter):
                if expect_data:
                    fields.append(
                        MultiPartFormDataField(name, content_type, b"\r\n".join(data).rstrip())
                    )
                    data.clear()

                expect_data = False
                continue

            if expect_data:
                data.append(line)
                continue

            if not line or line == b"\r\n":
                expect_data = True
                continue

            text = line.decode().lower()

            if text.startswith(CONTENT_DISPOSITION):
                name = get_name(text)
            elif text.startswith(CONTENT_TYPE):
                content_type = get_content_type(text)

        return fields



from __future__ import annotations

from abc import ABC
from abc import abstractmethod
from base64 import b64encode
import logging

from http.server import BaseHTTPRequestHandler


logger = logging.getLogger(__name__)


class AuthenticationError(Exception):
    pass


class Authenticator(ABC):

    @abstractmethod
    def authenticate(self, request: BaseHTTPRequestHandler):
        """Do the authentication."""


class BasicAuthenticator(Authenticator):

    @classmethod
    def _create_key(cls, username: str, password: str) -> str:
        return b64encode(f"{username}:{password}".encode()).decode()

    def __init__(self, username: str, password: str, realm: str = ""):
        self._key = self._create_key(username, password)
        self.realm = realm

    def authenticate(self, request: BaseHTTPRequestHandler):
        logger.debug("Using HTTP basic authentication.")

        authorization = request.headers.get("Authorization")

        if not authorization:
            self._handle_authentication_required(request)
            raise AuthenticationError("Authentication required.")

        if not self._is_authorized(authorization):
            self._handle_authentication_failed(request)
            raise AuthenticationError("Invalid username or password.")

        self._handle_authentication_succeeded(request)

    def _is_authorized(self, authorization: str) -> bool:
        return authorization.endswith(self._key)

    def _handle_authentication_required(self, request: BaseHTTPRequestHandler):
        logger.debug("HTTP basic authentication required.")

        request.send_response(401)
        request.send_header("WWW-Authenticate", f"Basic realm='{self.realm}'")
        request.send_header('Content-type', 'text/html')
        request.send_header("content-length", "0")
        request.end_headers()

    def _handle_authentication_failed(self, request: BaseHTTPRequestHandler):
        logger.debug("HTTP basic authentication failed.")

        request.send_response(401)
        request.send_header("WWW-Authenticate", f"Basic realm='{self.realm}'")
        request.send_header('Content-type', 'text/html')
        request.send_header("content-length", "0")
        request.end_headers()

    def _handle_authentication_succeeded(self, request: BaseHTTPRequestHandler):
        logger.debug("HTTP basic authentication successes.")



import typing
from typing import Optional
from typing import Any

"""
Example:
    {
        "endpoints": [
            {
                "name": "foo",
                "documentation": "Foo.\n\nTest docstring.",
                "parameters": [
                    {
                        "name": "a",
                        "annotation": "int"
                    },
                    {
                        "name": "b",
                        "annotation": "str",
                        "default": "foo"
                    }
                ],
                "return_annotation": "bool"
            }
        ],
        "address": {
            "hostname": "127.0.0.1",
            "port": 8000
        }
    }
"""


class ClientFactory:
    def __init__(self, definition: dict, class_name: Optional[str] = None):
        self._definition = definition
        self._client_class_name = class_name or "Client"

    def generate(self):
        address = self._definition.get("address", {})
        host = address.get("host", "localhost")
        port = address.get("port", 8000)

        c = Class(name=self._client_class_name)
        c.doc = self._definition.get("documentation", "")
        c.methods.append(self.generate_init_method((host, port)))

        endpoints = self._definition.get("endpoints", [])
        for endpoint in endpoints:
            name = endpoint.get("name")
            doc = endpoint.get("documentation")
            return_annotation = endpoint.get("return_annotation")

            m = Method(name=name, return_annotation=return_annotation, doc=doc)
            c.methods.append(m)
            for parameter in endpoint.get("parameters", []):
                p = Parameter(
                    parameter["name"], parameter["annotation"], parameter.get("default", None)
                )
                m.parameters.append(p)

        lines = ["from __future__ import annotations\n"]
        lines.append("import typing")
        lines.append("from typing import Optional\n")
        lines.append("from remotecall import BaseClient\n\n")
        lines.append(str(c))

        return "\n".join(lines)

    @classmethod
    def generate_init_method(cls, server_address):
        return (
            f"    def __init__(self, server_address={server_address}):\n"
            f"        super().__init__(server_address=server_address)\n"
        )


class Parameter:
    def __init__(self, name: str, annotation: str, default: Any):
        self.name = name
        self.annotation = annotation
        self.default = default

    def __str__(self):
        lines = [f"{self.name}: {self.annotation}"]

        if self.default is not None:
            lines.append(f" = {self.default}")

        return "".join(lines)


class Method:
    def __init__(self, name: str, return_annotation: str, doc: str = None):
        self.name = name
        self.return_annotation = return_annotation
        self.parameters = []
        self.indent = "    "
        self.doc = indent_doc(doc, self.indent * 2)

    def __str__(self):
        lines = [f"{self.indent}def {self.name}(self"]

        # Signature
        for parameter in self.parameters:
            lines.append(f", {parameter}")
        lines.append(")")

        # Return type
        if self.return_annotation:
            lines.append(f" -> {self.return_annotation}")

        lines.append(":\n")

        # Docstring
        if self.doc:
            lines.append(self.indent * 2)
            lines.append(f'"""{self.doc}\n')
            lines.append(self.indent * 2)
            lines.append('"""\n')

        # Method body
        lines.append(self.indent * 2)
        lines.append(f'return self.call("{self.name}"')

        for parameter in self.parameters:
            lines.append(f", {parameter.name}={parameter.name}")

        lines.append(")\n")

        return "".join(lines)


class Class:
    def __init__(self, name: str):
        self.name = name.capitalize()
        self.methods = []
        self.indent = ""
        self.doc = None

    def __str__(self):
        lines = [f"class {self.name}(BaseClient):"]

        if self.doc:
            lines.append(f'    """{self.doc}')
            lines.append('    """')

        for method in self.methods:
            lines.append(str(method))

        return "\n".join(lines)


def indent_doc(doc: str, indent: str) -> str:
    if not doc:
        return doc

    lines = doc.split("\n")
    for i, line in enumerate(lines[1:], 1):
        lines[i] = indent + line
    return "\n".join(lines)



# Client errors
BAD_REQUEST = 400
NOT_FOUND = 404
LENGTH_REQUIRED = 411
UNSUPPORTED_MEDIA_TYPE = 415
TYPE_ERROR = 452
INVALID_ARGUMENT = 453


# Server errors
SERVER_ERROR = 500
SERVICE_UNAVAILABLE = 503



from __future__ import annotations

import logging
import typing
from http.server import BaseHTTPRequestHandler

from .response import Response
from .authentication import authenticate
from .posthandler import PostHandler
from .gethandler import GetHandler
from .errors import ClientError
from .constants.headers import CONTENT_LENGTH
from .constants.headers import CONTENT_TYPE
from .constants.statuscodes import LENGTH_REQUIRED
from .constants.statuscodes import SERVER_ERROR

if typing.TYPE_CHECKING:
    from .server import Server


logger = logging.getLogger(__name__)


class HTTPRequestHandler(BaseHTTPRequestHandler):
    """HTTP request handler."""

    protocol_version = "HTTP/1.1"

    @property
    def path_component(self) -> str:
        path, _, _ = self.path.partition("?")
        return path.lstrip("/")

    @property
    def content_type(self) -> str:
        return self.headers[str(CONTENT_TYPE)]

    @property
    def content_length(self) -> int:
        """Return content length."""
        if CONTENT_LENGTH not in self.headers:
            raise ClientError("Content-length required", LENGTH_REQUIRED)
        return int(self.headers[CONTENT_LENGTH])

    # https://stackoverflow.com/questions/2617615/slow-python-http-server-on-localhost
    def address_string(self):
        host, _ = self.client_address[:2]
        # return socket.getfqdn(host)
        return host

    def get_server(self) -> Server:
        return self.server  # self.server is referring to http.server.BaseServer.

    def do_HEAD(self):  # pylint: disable=invalid-name
        logger.debug("Received HEAD request '%s'", self.path_component)
        self.send_response(200)
        self.send_header('Content-type', 'application/json; charset=utf-8')
        self.end_headers()
        logger.debug("Finished HEAD request '%s'", self.path_component)

    def do_GET(self):  # pylint: disable=invalid-name
        logger.debug("Received GET request '%s'", self.path_component)
        self._handle_get()
        logger.debug("Finished GET request '%s'", self.path_component)

    def do_POST(self):  # pylint: disable=invalid-name
        logger.debug("Received POST request '%s'", self.path_component)
        self._handle_post()
        logger.debug("Finished POST request '%s'", self.path_component)

    @authenticate
    def _handle_post(self):
        try:
            handler = PostHandler(self)
            response = handler.handle()
        except ClientError as err:
            logger.exception(err)
            response = self._create_error_response(message=err.message, status_code=err.status_code)
        except Exception as err:  # pylint: disable=broad-exception-caught
            logger.exception(err)

            # Handler is expected to identify expectable errors. This try-catch block catches
            # unexpected server errors that could not be identified.
            response = self._create_error_response(message=str(err), status_code=SERVER_ERROR)

        self._send_response(response)

    @authenticate
    def _handle_get(self):
        try:
            handler = GetHandler(self)
            response = handler.handle()
        except Exception as err:  # pylint: disable=broad-exception-caught
            logger.exception(err)
            response = self._create_error_response(message=str(err), status_code=SERVER_ERROR)

        self._send_response(response)

    def _send_response(self, response: Response):
        """Send response.
        """
        logger.debug("Sending response..")
        bytes_to_send = b"" if response.value is None else response.value

        try:
            self.send_response(response.status_code)

            if response.content_type:
                self.send_header(CONTENT_TYPE, response.content_type)

            self.send_header(CONTENT_LENGTH, str(len(bytes_to_send)))
            self.end_headers()
            self.wfile.write(bytes_to_send)
        except TypeError as err:
            logger.exception(err)
        except Exception as err:  # pylint: disable=broad-exception-caught
            logger.exception(err)

    def _create_error_response(self, message: str, status_code: int) -> Response:
        logger.error(message)
        logger.debug(str(self))

        server = self.get_server()
        codec = server.codecs.get_codec_by_value(message)
        return Response(codec.encode(message), codec.get_content_type(), status_code=status_code)

    def __str__(self):
        info = []
        info.append(f"{self.path}")
        info.append(f"{str(self.headers).strip()}")
        return "\n".join(info)



"""Image codec to be used with remotecall library.

Usage:
    from remotecall.extracodecs.imagecodec import ImageCodec
"""
import typing
import io
import base64

from PIL import Image

from remotecall import Codec


class ImageCodec(Codec):
    IMAGE_FORMAT = "png"

    @classmethod
    def is_matching_content_type(cls, content_type: str) -> bool:
        return content_type.startswith("application/image")

    @classmethod
    def is_matching_type(cls, type_: typing.Type) -> bool:
        return issubclass(type_, cls.get_type())

    @classmethod
    def get_type(cls) -> typing.Type:
        return Image.Image

    def get_content_type(self) -> str:
        return f"application/image-{self.IMAGE_FORMAT}"

    def encode(self, image: Image) -> bytes:
        buffer = io.BytesIO()
        image.save(buffer, format=self.IMAGE_FORMAT)
        return base64.b64encode(buffer.getvalue())

    def decode(self, data: bytes) -> Image:
        base64_str = io.BytesIO(base64.b64decode(data))
        return Image.open(base64_str)



import os
from setuptools import setup, find_packages

module_name = "remotecall"
meta = {}


with open(
    os.path.join(os.path.dirname(__file__), "src", module_name, "_meta.py"), "rt"
) as f:
    exec(f.read(), meta)  # pylint: disable=exec-used


setup(
    name=module_name,
    version=meta["__version__"],
    description="The module provides functionality to expose Python functions to be called remotely over ethernet.",
    long_description_content_type="text/markdown",
    long_description=open("README.md").read(),
    url="https://gitlab.com/slaine/{}.git/".format(module_name),
    author=meta["__author__"],
    author_email="sami.jy.laine@gmail.com",
    entry_points={
        "console_scripts": ["main=remotecall.__main__:app"],
    },
    license=meta["__license__"],
    packages=find_packages(where="src"),
    package_dir={"": "src"},
    zip_safe=False,
    include_package_data=True,
    install_requires=[
        "requests >= 2.28.1",
    ],
    extras_require={
        "test": ["pytest"],
        "full": ["numpy", "pillow"],
        "numpycodec": ["numpy"],
        "imagecodec": ["pillow"],
    },
)



__all__ = ["__author__", "__license__", "__version__", "__credits__", "__maintainer__"]
__author__ = "Sami Laine"
__license__ = "BSD-3-Clause"
__version__ = "0.1.5"
__credits__ = []
__maintainer__ = "Sami Laine"



from __future__ import annotations

import typing

from .response import JsonResponse

if typing.TYPE_CHECKING:
    from .requesthandler import HTTPRequestHandler


class GetHandler:
    def __init__(self, request: HTTPRequestHandler):
        self.request = request
        self._server = request.get_server()

    def handle(self) -> JsonResponse:  # pylint: disable=broad-exception-caught
        endpoints = [e.to_dict() for e in self._server.endpoints.values()]
        definition = {
            "endpoints": endpoints,
            "ssl_enabled": self._server.ssl_enabled
        }

        return JsonResponse(definition)



from __future__ import annotations

import logging
import typing

from .authenticator import Authenticator
from .authenticator import BasicAuthenticator
from .authenticator import AuthenticationError

if typing.TYPE_CHECKING:
    from ..requesthandler import HTTPRequestHandler


logger = logging.getLogger(__name__)


def authenticate(function):
    def wrapper(request: HTTPRequestHandler, *args, **kwargs):
        server = request.get_server()
        authenticator_ = server.get_authenticator()

        if authenticator_:
            try:
                authenticator_.authenticate(request)
            except AuthenticationError as err:
                logger.warning(err)
                return

        function(request, *args, **kwargs)

    return wrapper



import json


class Response:
    def __init__(self, value, content_type: str, status_code: int = 200):
        self._value = value
        self.content_type = content_type
        self.status_code = status_code

    @property
    def value(self):
        return self._value


class JsonResponse(Response):
    def __init__(self, value, **kwargs):
        json_object = json.dumps(value)
        super().__init__(
            value=json_object.encode("utf-8"),
            content_type="application/json; charset=utf-8",
            **kwargs
        )


class ErrorResponse(Response):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, content_type="text/html", **kwargs)



from __future__ import annotations

import abc
import json
from typing import Type


class NotFoundError(Exception):
    pass


class Codecs:
    def __init__(self, codecs: list[Type[Codec]]):
        self._codecs = codecs

    def register(self, codec: Type[Codec]):
        """Register codec type."""
        if not issubclass(codec, Codec):
            raise ValueError(f"Expecting type subclassed from Codec. Got {type(codec)}.")

        if codec in self._codecs:
            return

        self._codecs.append(codec)

    def unregister(self, codec: Type[Codec]):
        """Unregister codec type."""
        self._codecs.pop(codec, None)

    def clear(self):
        """Clear all registered codec types."""
        self._codecs.clear()

    def get_codec_by_value(self, obj) -> Codec:
        """Get codec by Python object."""
        for codec in self._codecs:
            if codec.is_matching_type(type(obj)):
                return codec.create_by_value(obj)

        raise NotFoundError(f"No matching codec found for object: {obj}")

    def get_codec_by_type(self, type_: Type) -> Codec:
        """Create codec by Python type."""
        for codec in self._codecs:
            if codec.is_matching_type(type_):
                return codec()

        raise NotFoundError(f"No matching codec found for type: {type_}")

    def get_codec_by_content_type(self, content_type: str) -> Codec:
        """Create codec by content-type."""
        for codec in self._codecs:
            if codec.is_matching_content_type(content_type):
                return codec.create_by_content_type(content_type)

        raise NotFoundError(f"No matching codec found for content-type: {content_type}")


class Codec(abc.ABC):
    subclasses = []

    def __init_subclass__(cls, **kwargs):
        Codec.subclasses.append(cls)

    @classmethod
    def create_by_value(cls, obj) -> Codec:
        """Create codec by Python object."""
        return cls()

    @classmethod
    def create_by_content_type(cls, content_type: str) -> Codec:
        """Create codec by content type."""
        return cls()

    @classmethod
    def is_matching_type(cls, type_: Type) -> bool:
        """Is codec matching with the given Python type."""
        return type_ == cls.get_type()

    @classmethod
    def is_matching_content_type(cls, content_type: str) -> bool:
        """Is codec matching with content-type."""
        type_name = cls.get_type().__name__
        return content_type == f"application/{type_name}"

    def get_content_type(self) -> str:
        """Get content-type associated with codec."""
        return f"application/{self.get_type().__name__}"

    @classmethod
    @abc.abstractmethod
    def get_type(cls) -> Type:
        """Get Python type associated with codec."""

    @abc.abstractmethod
    def encode(self, obj) -> bytes:
        """Encode object as bytes."""

    @abc.abstractmethod
    def decode(self, data: bytes) -> object:
        """Decode bytes as object."""


class NoneCodec(Codec):
    CONTENT_TYPE = "application/none"

    @classmethod
    def get_type(cls) -> Type:
        return type(None)

    @classmethod
    def is_matching_content_type(cls, content_type: str) -> bool:
        return content_type == cls.CONTENT_TYPE

    def get_content_type(self) -> str:
        return self.CONTENT_TYPE

    def encode(self, obj: None) -> bytes:
        return "None".encode()

    def decode(self, data: bytes) -> None:
        return None


class BytesCodec(Codec):

    @classmethod
    def get_type(cls) -> Type:
        return bytes

    def encode(self, obj: bytes) -> bytes:
        return obj

    def decode(self, data: bytes) -> bytes:
        return data


class IntCodec(Codec):

    @classmethod
    def get_type(cls) -> Type:
        return int

    def encode(self, obj: int) -> bytes:
        return str(obj).encode()

    def decode(self, data: bytes) -> int:
        return int(data)


class FloatCodec(Codec):

    @classmethod
    def get_type(cls) -> Type:
        return float

    def encode(self, obj: float) -> bytes:
        return str(obj).encode()

    def decode(self, data: bytes) -> float:
        return float(data)


class BoolCodec(Codec):

    @classmethod
    def get_type(cls) -> Type:
        return bool

    def encode(self, obj: bool) -> bytes:
        return str(obj).encode()

    def decode(self, data: bytes) -> bool:
        str_value = data.decode()
        return str_value.lower() == "true"


class StrCodec(Codec):

    @classmethod
    def get_type(cls) -> Type:
        return str

    def encode(self, obj: str) -> bytes:
        return obj.encode()

    def decode(self, data: bytes) -> str:
        return data.decode()


class DictCodec(Codec):

    @classmethod
    def get_type(cls) -> Type:
        return dict

    def encode(self, obj: dict) -> bytes:
        return json.dumps(obj).encode()

    def decode(self, data: bytes) -> dict:
        return json.loads(data.decode())


class ListCodec(Codec):

    @classmethod
    def get_type(cls) -> Type:
        return list

    def encode(self, obj: list) -> bytes:
        return json.dumps(obj).encode()

    def decode(self, data: bytes) -> list:
        return json.loads(data.decode())


class TupleCodec(Codec):

    @classmethod
    def get_type(cls) -> Type:
        return tuple

    def encode(self, obj: tuple) -> bytes:
        return json.dumps(obj).encode()

    def decode(self, data: bytes) -> tuple:
        return tuple(json.loads(data.decode()))
